#include "kem_int_flows.h"
#include "sampling.h"
#include "parallel_hash.h"
#include "gf2x.h"

ret_t encrypt(OUT ct_t *ct,
              IN const pk_t *pk,
              IN const split_e_t *splitted_e)
{
    //Pad the public key
    dbl_pad_ct_t p_ct;
    pad_pk_t p_pk = {0};
    VAL(p_pk) = *pk;

    GUARD(cyclic_product(VAL(p_ct).raw, PTRV(splitted_e)[1].raw, VAL(p_pk).raw));
    GUARD(gf2x_add(VAL(p_ct).raw, VAL(p_ct).raw, PTRV(splitted_e)[0].raw, R_SIZE));
    
    *ct = VAL(p_ct);

    EDMSG("c:  "); print((uint64_t*)ct->raw, R_BITS);

    return SUCCESS;
}

ret_t calc_pk(OUT pk_t *pk, IN OUT sk_t *sk)
{
    DEFER_CLEANUP(r_t inv, r_cleanup);
    
    GUARD(mod_inv(inv.raw, PTR(sk).bin[0].raw));
    GUARD(cyclic_product(pk->raw, PTR(sk).bin[1].raw, inv.raw));

    PTR(sk).pk = *pk;

    EDMSG("g0: "); print((uint64_t*)pk->raw, R_BITS);
    return SUCCESS;
}

// Generate the Shared Secret (K(e))
void get_ss(OUT ss_t *out, IN const e_t *e)
{
    DMSG("    Enter get_ss.\n");

    // Calculate the hash
    sha_hash_t hash = {0};
    parallel_hash(&hash, e->raw, sizeof(*e));

    // Truncate the final hash into K by copying only the LSBs
    memcpy(out->raw, hash.u.raw, sizeof(*out));

    secure_clean(hash.u.raw, sizeof(hash));
    DMSG("    Exit get_ss.\n");
}


#if BATCH_SIZE > 1
ret_t init_batch(IN OUT aes_ctr_prf_state_t *h_prf_state,
                 IN OUT batch_ctx_t *ctx)
{
    if(ctx->cnt != 0)
    {
        //Already initialized.
        return SUCCESS;
    }

    //First time - Init ad calc.
    for(uint32_t i=0 ; i < BATCH_SIZE ; ++i)
    {
        GUARD(generate_sparse_fake_rep(ctx->h0[i].u.qw, 
                                       ctx->sk_wlist[i].val,
                                       sizeof(VAL(ctx->h0[i])),
                                       h_prf_state));
    }

    ctx->tmp0[0] = VAL(ctx->h0[0]);

    for(uint32_t i = 1 ; i < BATCH_SIZE ; ++i)
    {
        GUARD(cyclic_product(ctx->tmp0[i].raw, 
                             ctx->tmp0[i-1].raw, 
                             VAL(ctx->h0[i]).raw));
    }

    GUARD(mod_inv(ctx->tmp1[BATCH_SIZE-1].raw, ctx->tmp0[BATCH_SIZE-1].raw));

    for(uint32_t i = (BATCH_SIZE - 2) ; i >= 1 ; i--)
    {
        GUARD(cyclic_product(ctx->tmp1[i].raw, 
                             ctx->tmp1[i+1].raw, 
                             VAL(ctx->h0[i+1]).raw));
    }

    return SUCCESS;
}

ret_t get_batch_keys(OUT pk_t *pk, 
                     IN OUT sk_t *sk, 
                     IN OUT batch_ctx_t *ctx)
{
    //Set the private key.
    PTR(sk).bin[0] = VAL(ctx->h0[ctx->cnt]);
    PTR(sk).wlist[0] = ctx->sk_wlist[ctx->cnt];

    if(ctx->cnt == 0)
    {
        GUARD(cyclic_product(ctx->inv.raw, 
                             ctx->tmp1[1].raw, 
                             VAL(ctx->h0[1]).raw));
    }
    else
    {
        GUARD(cyclic_product(ctx->inv.raw, 
                             ctx->tmp1[ctx->cnt].raw, 
                             ctx->tmp0[ctx->cnt-1].raw));
    }
   
    //Set the public key.
    GUARD(cyclic_product(pk->raw, PTR(sk).bin[1].raw, ctx->inv.raw));

    EDMSG("g0: "); print((uint64_t*)pk->raw, R_BITS);
    
    //Increment the counter
    ctx->cnt = (ctx->cnt+1) % BATCH_SIZE;

    return SUCCESS;
}

#endif
