#include "kem_int_flows.h"
#include "sampling.h"
#include "parallel_hash.h"
#include "gf2x.h"

#ifdef BIKE3_BANDWIDTH_OPT
ret_t encrypt(OUT ct_t *ct,
              IN const pk_t *pk,
              IN const split_e_t *splitted_e,
              IN const padded_r_t *e_extra)
{
    DEFER_CLEANUP(padded_r_t g = {0}, padded_r_cleanup);
    GUARD(sample_uniform_r_bits(VAL(g).raw, &PTR(pk).pk1, MUST_BE_ODD));

    // ct = (e1*pk0 + e_extra, e1*pk1 + e0)
    GUARD(cyclic_product(PTRV(ct)[0].raw, PTRV(splitted_e)[1].raw, PTR(pk).pk0.raw));
    GUARD(cyclic_product(PTRV(ct)[1].raw, PTRV(splitted_e)[1].raw, VAL(g).raw));
    GUARD(gf2x_add(PTRV(ct)[0].raw, PTRV(ct)[0].raw, PTRV(e_extra).raw, R_SIZE));
    GUARD(gf2x_add(PTRV(ct)[1].raw, PTRV(ct)[1].raw, PTRV(splitted_e)[0].raw, R_SIZE));

    EDMSG("c0: "); print((uint64_t*)PTRV(ct)[0].raw, R_BITS);
    EDMSG("c1: "); print((uint64_t*)PTRV(ct)[1].raw, R_BITS);

    return SUCCESS;
}

ret_t calc_pk(OUT pk_t *pk,
              IN const seed_t *g_seed,
              IN OUT sk_t *sk)
{
    // pk = (h1 + g*h0, g)
    DEFER_CLEANUP(padded_r_t g = {0}, padded_r_cleanup);
    DEFER_CLEANUP(r_t tmp1, r_cleanup);

    GUARD(sample_uniform_r_bits(VAL(g).raw, g_seed, MUST_BE_ODD));

    GUARD(cyclic_product(tmp1.raw, VAL(g).raw, PTR(sk).bin[0].raw));
    GUARD(gf2x_add(PTR(pk).pk0.raw, tmp1.raw, PTR(sk).bin[1].raw, R_SIZE));

    // Copy g to pk[1]
    PTR(pk).pk1 = *g_seed;
    
    // Store the pk for later use
    PTR(sk).pk = *pk;
    
    EDMSG("g0: "); print((uint64_t*)PTRV(pk)[0].raw, R_BITS);
    EDMSG("g1: "); print((uint64_t*)PTRV(pk)[1].raw, R_BITS);

    return SUCCESS;
}

#else

ret_t encrypt(OUT ct_t *ct,
              IN const pk_t *pk,
              IN const split_e_t *splitted_e,
              IN const padded_r_t *e_extra)
{
    // ct = (e1*pk0 + e_extra, e1*pk1 + e0)
    GUARD(cyclic_product(PTRV(ct)[0].raw, PTRV(splitted_e)[1].raw, PTRV(pk)[0].raw));
    GUARD(cyclic_product(PTRV(ct)[1].raw, PTRV(splitted_e)[1].raw, PTRV(pk)[1].raw));
    GUARD(gf2x_add(PTRV(ct)[0].raw, PTRV(ct)[0].raw, PTRV(e_extra).raw, R_SIZE));
    GUARD(gf2x_add(PTRV(ct)[1].raw, PTRV(ct)[1].raw, PTRV(splitted_e)[0].raw, R_SIZE));

    EDMSG("c0: "); print((uint64_t*)PTRV(ct)[0].raw, R_BITS);
    EDMSG("c1: "); print((uint64_t*)PTRV(ct)[1].raw, R_BITS);

    return SUCCESS;
}

ret_t calc_pk(OUT pk_t *pk,
              IN const seed_t *g_seed,
              IN OUT sk_t *sk)
{
    // pk = (h1 + g*h0, g)
    DEFER_CLEANUP(padded_r_t g = {0}, padded_r_cleanup);
    DEFER_CLEANUP(r_t tmp1, r_cleanup);

    GUARD(sample_uniform_r_bits(VAL(g).raw, g_seed, MUST_BE_ODD));

    GUARD(cyclic_product(tmp1.raw, VAL(g).raw, PTR(sk).bin[0].raw));
    GUARD(gf2x_add(PTRV(pk)[0].raw, tmp1.raw, PTR(sk).bin[1].raw, R_SIZE));

    // Copy g to pk[1]
    PTRV(pk)[1] = VAL(g);
    
    // Store the pk for later use
    PTR(sk).pk = *pk;
    
    EDMSG("g0: "); print((uint64_t*)PTRV(pk)[0].raw, R_BITS);
    EDMSG("g1: "); print((uint64_t*)PTRV(pk)[1].raw, R_BITS);

    return SUCCESS;
}
#endif

// Generate the Shared Secret (K(e))
void get_ss(OUT ss_t *out, IN const e_t *e, IN const r_t *e_extra)
{
    DMSG("    Enter get_ss.\n");

    uint8_t tmp[N_SIZE + R_SIZE];
    memcpy(tmp, e->raw, sizeof(e->raw));
    memcpy(tmp + sizeof(e->raw), e_extra->raw, sizeof(e_extra->raw));

    // Calculate the hash
    sha_hash_t hash = {0};
    parallel_hash(&hash, tmp, sizeof(tmp));

    // Truncate the final hash into K by copying only the LSBs
    memcpy(out->raw, hash.u.raw, sizeof(*out));

    secure_clean(hash.u.raw, sizeof(hash));
    DMSG("    Exit get_ss.\n");
}
