/***************************************************************************
* Additional implementation of "BIKE: Bit Flipping Key Encapsulation". 
* Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
*
* Written by Nir Drucker and Shay Gueron
* AWS Cryptographic Algorithms Group
* (ndrucker@amazon.com, gueron@amazon.com)
*
* The license is detailed in the file LICENSE.md, and applies to this file.
* ***************************************************************************/

#include "kem.h"
#include "kem_int_flows.h"
#include "sampling.h"
#include "decode.h"

#if BIKE_VER==3
// convert a sequence of uint8_t elements which uses just a single bit per 
// byte (either 0 or 1) to a sequence of uint8_t which fully uses all 8-bits 
// of an uint8_t element.
_INLINE_ void convert_from_redundant_rep(OUT uint8_t* out, 
                                         IN const uint8_t *in, 
                                         IN const uint64_t len)
{
    bike_static_assert((len % 8 != 0), from_red_err);

    const uint32_t num_of_bytes = (len/8) + 1;
    uint32_t ctr = 0;

    for (uint32_t i = 0; i < (num_of_bytes - 1); i++)
    {
        for (uint32_t j = 0; j < 8; j++)
        {
            out[i] |= (uint8_t)(in[ctr++] << j);
        }
    }

    for (uint32_t j = 0; ctr < len; ctr++, j++)
    {
        out[num_of_bytes-1] |= (uint8_t)(in[ctr] << j);
    }
}

#endif

////////////////////////////////////////////////////////////////
//The three APIs below (keygeneration, encapsulate, decapsulate) are defined by NIST:
//In addition there are two KAT versions of this API as defined.
////////////////////////////////////////////////////////////////
int crypto_kem_keypair(OUT unsigned char *pk, OUT unsigned char *sk)
{
    // Convert to this implementation types
    sk_t *l_sk = (sk_t *)sk;
    pk_t *l_pk = (pk_t *)pk;

    // For DRBG and AES_PRF
    DEFER_CLEANUP(double_seed_t seeds = {0}, double_seed_cleanup);
    DEFER_CLEANUP(aes_ctr_prf_state_t h_prf_state = {0}, aes_ctr_prf_state_cleanup);
    // Padded for internal use
    // We don't want to send the padded data outside to save BW.
    DEFER_CLEANUP(pad_sk_t p_sk = {0}, pad_sk_cleanup);
    
    // Get the entrophy seeds
    get_seeds(&seeds, KEYGEN_SEEDS);

    DMSG("  Enter crypto_kem_keypair.\n");
    DMSG("    Calculating the secret key.\n");

    // Both h0 and h1 use the same context
    GUARD(init_aes_ctr_prf_state(&h_prf_state, MAX_AES_INVOKATION, &seeds.u.v.s1));
    
    // Make sure that the wlists are zeroed for the KATs.
    memset(l_sk, 0, sizeof(sk_t));

#if BATCH_SIZE > 1
    static batch_ctx_t batch_ctx = {.cnt = 0};
    GUARD(init_batch(&h_prf_state, &batch_ctx));
#else
    GUARD(generate_sparse_fake_rep(p_sk[0].u.qw, PTR(l_sk).wlist[0].val,
                                   sizeof(p_sk[0]), &h_prf_state));
    // Copy data
    PTR(l_sk).bin[0] = VAL(p_sk[0]);
#endif
    GUARD(generate_sparse_fake_rep(p_sk[1].u.qw, PTR(l_sk).wlist[1].val,
                                   sizeof(p_sk[1]), &h_prf_state));
    // Copy data
    PTR(l_sk).bin[1] = VAL(p_sk[1]);

    DMSG("    Calculating the public key.\n");

#if BIKE_VER==1
    GUARD(calc_pk(l_pk, &seeds.u.v.s2, p_sk));
#elif (BIKE_VER == 2)
  #if (BATCH_SIZE > 1)
    GUARD(get_batch_keys(l_pk, l_sk, &batch_ctx));
  #else
    GUARD(calc_pk(l_pk, l_sk));
  #endif
#elif (BIKE_VER == 3)
    GUARD(calc_pk(l_pk, &seeds.u.v.s2, l_sk));
#endif

    EDMSG("h0: "); print((uint64_t*)&PTR(l_sk).bin[0], R_BITS);
    EDMSG("h1: "); print((uint64_t*)&PTR(l_sk).bin[1], R_BITS);
    EDMSG("h0c:"); print((uint64_t*)&PTR(l_sk).wlist[0], SIZEOF_BITS(compressed_idx_dv_t));
    EDMSG("h1c:"); print((uint64_t*)&PTR(l_sk).wlist[1], SIZEOF_BITS(compressed_idx_dv_t));
    DMSG("  Exit crypto_kem_keypair.\n");

    return SUCCESS;
}

// Encapsulate - pk is the public key,
//               ct is a key encapsulation message (ciphertext),
//               ss is the shared secret.
int crypto_kem_enc(OUT unsigned char *ct,
                   OUT unsigned char *ss,
                   IN  const unsigned char *pk)
{
    DMSG("  Enter crypto_kem_enc.\n");

    // Convert to this implementation types
    const pk_t *l_pk = (const pk_t *)pk;
    ct_t *l_ct = (ct_t *)ct;
    ss_t *l_ss = (ss_t *)ss;
    DEFER_CLEANUP(padded_e_t e = {0}, padded_e_cleanup);

    // For NIST DRBG_CTR
    DEFER_CLEANUP(double_seed_t seeds = {0}, double_seed_cleanup);
    DEFER_CLEANUP(aes_ctr_prf_state_t e_prf_state = {0}, aes_ctr_prf_state_cleanup);

    // Get the entrophy seeds
    get_seeds(&seeds, ENCAPS_SEEDS);

    // Random data generator
    // Using first seed
    GUARD(init_aes_ctr_prf_state(&e_prf_state, MAX_AES_INVOKATION, &seeds.u.v.s1));

    DMSG("    Generating error.\n");
    compressed_idx_t_t dummy;
    GUARD(generate_sparse_rep(e.u.qw, dummy.val, T1, N_BITS, sizeof(e), &e_prf_state));

    EDMSG("e:  "); print((uint64_t*)VAL(e).raw, sizeof(e)*8);

    // Split e into e0 and e1. Initialization is done in split_e
    DEFER_CLEANUP(split_e_t splitted_e, split_e_cleanup);
    split_e(&splitted_e, &VAL(e));
    
    EDMSG("e0: "); print((uint64_t*)VAL(splitted_e)[0].raw, R_BITS);
    EDMSG("e1: "); print((uint64_t*)VAL(splitted_e)[1].raw, R_BITS);

#if BIKE_VER==3
    DEFER_CLEANUP(padded_r_t e_extra, padded_r_cleanup);

    GUARD(generate_sparse_rep(e_extra.u.qw, dummy.val, T1/2, 
                              R_BITS, sizeof(e_extra), &e_prf_state));
    EDMSG("e_extra: "); print((uint64_t*)e_extra.u.raw, R_BITS);
#endif

    // Computing ct = enc(pk, e)
    // Using second seed
    DMSG("    Encrypting.\n");
#if   BIKE_VER==1
    GUARD(encrypt(l_ct, l_pk, &seeds.u.v.s2, &splitted_e));
#elif BIKE_VER==2
    GUARD(encrypt(l_ct, l_pk, &splitted_e));
#elif BIKE_VER==3
    GUARD(encrypt(l_ct, l_pk, &splitted_e, &e_extra));
#endif

    DMSG("    Generating shared secret.\n");

#if BIKE_VER==3
    get_ss(l_ss, &VAL(e), &VAL(e_extra));
#else
    get_ss(l_ss, &VAL(e));
#endif

    EDMSG("ss: "); print((uint64_t*)l_ss->raw, SIZEOF_BITS(*l_ss));
    DMSG("  Exit crypto_kem_enc.\n");
    return SUCCESS;
}

// Decapsulate - ct is a key encapsulation message (ciphertext),
//               sk is the private key,
//               ss is the shared secret
int crypto_kem_dec(OUT unsigned char *ss,
                   IN const unsigned char *ct,
                   IN const unsigned char *sk)
{
    DMSG("  Enter crypto_kem_dec.\n");

    // Convert to this implementation types
    const sk_t *l_sk = (const sk_t *)sk;
    const ct_t *l_ct = (const ct_t *)ct;
    ss_t *l_ss = (ss_t *)ss;
    
    // Force zero initialization
    DEFER_CLEANUP(syndrome_t syndrome = {0}, syndrome_cleanup);
    DEFER_CLEANUP(e_t e = {0}, e_cleanup);
    
    DMSG("  Computing s.\n");
    GUARD(compute_syndrome(&syndrome, l_ct, l_sk));

    DMSG("  Decoding.\n");
    GUARD(decode(&e, &syndrome, l_ct, l_sk, U_ERR));

    // Check if the error weight equals T1
    if (count_ones(e.raw, sizeof(e)) != T1)
    {
        MSG("    Error weight is not t\n");
        BIKE_ERROR(E_ERROR_WEIGHT_IS_NOT_T);
    }

#if (BIKE_VER==1) || (BIKE_VER==2)
    get_ss(l_ss, &e);
#else
    DEFER_CLEANUP(r_t e_extra = {0}, r_cleanup);
    convert_from_redundant_rep(e_extra.raw, syndrome.u.v.dup1.raw, R_BITS);

    if (count_ones(e_extra.raw, sizeof(e_extra.raw)) != (T1/2))
    {
        MSG("    Extra error weight is not t/2\n");
        BIKE_ERROR(E_ERROR_WEIGHT_IS_NOT_T);
    }

    EDMSG("e_extra: "); print((uint64_t*)e_extra.raw, R_BITS);

    get_ss(l_ss, &e, &e_extra);
#endif

    DMSG("  Exit crypto_kem_dec.\n");
    return SUCCESS;
}
